function [alphaOpt,bOpt,eta,omegaSquared,finalLpDualObj,theta,zeta,aTQa_t1t2] = ...
    mirrorDescentWrapper(K_tr,alpha_smo,y,nodeIdx,C,rho,eta,data,dataset,inverted,mu,varargin)

tube_eps = .1;
myRidge = 1e-8;
args = varargin;
nargs = length(args);
theta_flag = true;
for i=1:2:nargs
    switch args{i},
        case 'final', theta_flag = true;
    end
end

% solve the milasso for a given regularization parameter
mltask = dataset.mltask;
numTrain = dataset.numTrain;
T = length(numTrain);
p =length(data.affinity);
affs = data.affinity;
weights = data.weights;
hull = data.hull;
if inverted
    hull = 3-hull;
end
singleTaskMuFactor = 1+1/mu;

n = sum(numTrain);
fprintf('\n[mirrorDescentWrapper] No. of nodes: %d\n',p);
Y = [];
for t=1:T
    Y = [Y;y{t}];
end

lambda = 1/C;
%Calculate Kernel
myEps = 1e-6;
% READ OPTIONAL PARAMETERS

tic;
if (p >= 1)
    if (rho ~= 2)
        rhobar = rho/(2-rho);% MKL norm in primal
        rhobarstar = rhobar/(rhobar-1);
    else
        rhobar = inf;% MKL norm in primal
        rhobarstar = 1;
    end
    ITER_MAX = 500;

%     % WARM START
    if isempty(eta)
        eta = ones(p,1)/p;
    elseif ~all(eta)
        eta = eta + 1e-8;
    end
    
    eta = eta/sum(eta);

    % MIRROR DESCENT
    errCount = 0;
    dualObj = 1.7977e+300;
    min_alpha = nan;
    min_b = nan;
    min_eta = nan;
    min_zeta = nan;
    min_dualObj = 1.7977e+300;
    min_delta = nan;
    min_aTQa_smo = nan;
    
    for iter_gamma = 1:ITER_MAX
        % Solve for alpha ;given eta
        dualObjOld = dualObj;
        if (sum(isnan(eta)))
            fprintf('eta is NAN\n');
            keyboard;
        end
        fprintf('.');
        zetaFake = calculateZetaFake(affs,rho,weights,eta);
        zeta = nthroot(zetaFake,rhobarstar);
        if (rho ~= 2)
            tic
            fprintf('@');
            L_p = rhobar;
            switch (mltask)
                case {'classification'}
                    %% SMO for classification
                    if iter_gamma == 1
                        [alpha_smo,dualObj,aTQa_smo,bOpt] = smo_lpMKL_multiTask_K_mu_MKL(K_tr,nodeIdx,zeta,hull,Y,abs(alpha_smo),1/lambda,rhobar,dataset,mu);
                    else
                        [alpha_smo,dualObj,aTQa_smo,bOpt] = smo_lpMKL_multiTask_K_mu_MKL(K_tr,nodeIdx,zeta,hull,Y,alpha_smo,1/lambda,rhobar,dataset,mu);
                    end
                    alphaOpt = alpha_smo.*Y;
                    delta = 2*(sum(abs(alphaOpt)) - dualObj);
                     %% SMO for classification
                case {'regression'}
                    %% eps-SVR-SMO
                    [alpha_smo,dualObj,aTQa_smo,bOpt] = smo_lpMKL_reg_multiTask_K_mu_MKL(K_tr,nodeIdx,zeta,hull,Y,alpha_smo,1/lambda,L_p,dataset,tube_eps,mu);
                    alphaOpt = alpha_smo;
                    delta = 2*(Y'*alphaOpt - tube_eps*sum(abs(alphaOpt)) - dualObj);
                    %% eps-SVR-SMO
                otherwise
                    fprintf('\n[mirrorDescentWrapper] Unknown mltask type\n');
                    keyboard;
            end
            fprintf('@');
            toc
            fprintf('\nDualObj: %f\n',dualObj);
%             keyboard;
        else
            % for rho=2, only one effective kernel is needed
            fprintf('rho=2,p=2 NOT implemented!!\n');
            keyboard;
            
            k = T*(T+1)/2;
            coeff = zeros(k,1);
            for i=1:p
                presentTasks = find(hull(i,:)==2);
                if (length(presentTasks) == 1)
                    coeff(presentTasks) = coeff(presentTasks) + zeta(i)/singleTaskMuFactor;
                else
                    pairs = combntns(presentTasks,2);
                    index = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
                    coeff([presentTasks';index]) = coeff([presentTasks';index]) + zeta(i);
                end
            end
            K2 = zeros(n,n);
            for t1=1:T
                for t2=t1:T
                    if t1 == t2
                        index3 = t1;
                        index1 = (sum(numTrain(1:(t1-1)))+1):sum(numTrain(1:t1));
                        K2(index1,index1) = K2(index1,index1) + sum(K_tr(index1,index1,:),3)*coeff(index3);
                    else
                        index3 = T + (t1-1)*(T-t1/2) + t2 - t1;
                        index1 = (sum(numTrain(1:(t1-1)))+1):sum(numTrain(1:t1));
                        index2 = (sum(numTrain(1:(t2-1)))+1):sum(numTrain(1:t2));
                        Ktemp = sum(K_tr(index1,index2,:),3)*coeff(index3);
                        K2(index1,index2) = K2(index1,index2) + Ktemp;
                        K2(index2,index1) = K2(index2,index1) + Ktemp';
                    end
                end
            end
            
            nodeIdx_temp = true(n,1);% do not use ones(n,1)
            hull_temp = 2*ones(1,T);
            zeta_temp = 1;
            Lp_temp = 2;
            tic
            fprintf('@');
            switch (mltask)
                case {'classification'}
                    %% SMO for classification
                    if iter_gamma == 1
                        [alpha_smo,dualObj,aTQa_smo_fake,bOpt] = smo_lpMKL_multiTask_K_mu(K2,nodeIdx_temp,zeta_temp,hull_temp,Y,abs(alpha_smo),1/lambda,Lp_temp,dataset,0,mu);
                    else
                        [alpha_smo,dualObj,aTQa_smo_fake,bOpt] = smo_lpMKL_multiTask_K_mu(K2,nodeIdx_temp,zeta_temp,hull_temp,Y,alpha_smo,1/lambda,Lp_temp,dataset,0,mu);
                    end
                    alphaOpt = alpha_smo.*Y;
                    delta = 2*(sum(abs(alphaOpt)) - dualObj);
                     %% SMO for classification
                case {'regression'}
                    %% eps-SVR-SMO
                    [alpha_smo,dualObj,aTQa_smo_fake,bOpt] = smo_lpMKL_reg_multiTask_K_mu(K2,nodeIdx_temp,zeta_temp,hull_temp,Y,alpha_smo,1/lambda,Lp_temp,dataset,tube_eps,mu);
                    alphaOpt = alpha_smo;
                    delta = 2*(Y'*alphaOpt - tube_eps*sum(abs(alphaOpt)) - dualObj);
                    %% eps-SVR-SMO
                otherwise
                    fprintf('\n[mirrorDescentWrapper] Unknown mltask type\n');
                    keyboard;
            end
            fprintf('@');
            toc
            fprintf('\nDualObj: %f\n',dualObj);
            aTQa_smo = [];
        end

        % compute a_iY_i Kij Y_ja_j for all task i,j
        aTQa_t1t2 = zeros(T*(T+1)/2,1);
        for t1=1:T
            for t2 = t1:T
                index1 = (sum(numTrain(1:(t1-1)))+1):sum(numTrain(1:t1));
                index2 = (sum(numTrain(1:(t2-1)))+1):sum(numTrain(1:t2));
                if t1 == t2
                    index3 = t1;
                    aTQa_t1t2(index3) = alphaOpt(index1)'*sum(K_tr(index1,index2,:),3)*alphaOpt(index2);
                else
                    index3 = T + (t1-1)*(T-t1/2) + t2 - t1;
                    aTQa_t1t2(index3) = 2*(alphaOpt(index1)'*sum(K_tr(index1,index2,:),3)*alphaOpt(index2));
                end
            end
        end
        
        if (delta <= 0)
            fprintf('[mirrorDescentWrapper] Delta is negative\n');
            keyboard;
        end
        if ( 1e-6 <= (dualObj-dualObjOld)/dualObj)
            fprintf('[mirrorDescentWrapper] This is bad! dualObjOld=%e, dualObj=%e, rel.gap=%e\n',dualObjOld, dualObj,(dualObj-dualObjOld)/dualObj);
            if errCount > 5
                fprintf('[mirrorDescentWrapper] erCount exceeds limit of 5. breaking MD\n');
                break;
            else
                errCount = errCount+1;
            end
        else
            errCount = 0;
        end
        %book-keeping
        if (dualObj < min_dualObj)
            min_alpha = alphaOpt;
            min_b = bOpt;
            min_eta = eta;
            min_zeta = zeta;
            min_dualObj = dualObj;
            min_delta = delta;
            min_aTQa_smo = aTQa_smo;
            min_aTQa_t1t2 = aTQa_t1t2;
        end
        %book-keeping
        if (abs((dualObjOld-dualObj)/dualObj) < myEps)
            break;
        end
        
        T2 = delta^rhobarstar;
        [gradHelp flag] = calculateSumForDesc(aTQa_smo,affs,rho,weights,eta,repmat(zetaFake,size(K_tr,3),1));
        if flag
            fprintf('[mirrorDescentWrapper] flag of [calculateSumForDesc] is true\n');
            keyboard;
        end
        gradF = -(delta/T2)/(2*rhobarstar)*gradHelp;%alphaOpt,affs,rho,weights,eta,K,zetaFake,hull,T
        step = sqrt(log(p))/(sqrt(iter_gamma)*norm(gradF,inf));
        pvec = 1+log(eta) - step*gradF;
        eta = exp(pvec);
        eta = eta/sum(eta);
    end

    if (iter_gamma == ITER_MAX)
        fprintf('[mirrorDescentWrapper] *****Before convergence termination of MD*****, diff = %e\n',abs((dualObjOld-dualObj)/dualObj));
    end
    
    alphaOpt = min_alpha;
    bOpt = min_b;
    eta  = min_eta;
    zeta = min_zeta;
    dualObj = min_dualObj;
    delta = min_delta;
    aTQa_smo = min_aTQa_smo;
    aTQa_t1t2 = min_aTQa_t1t2;
    
    finalLpDualObj = dualObj;
    % Theta Computation
    if theta_flag % WRONG - debug for rho~=2 (zeta missing)
        theta = (aTQa_smo/delta).^(1/(rhobar-1));
    else
        theta = [];
    end
    omegaSquared = delta;
end

fprintf('[mirrorDescentWrapper] MDiterations: %d, MDtime: ',iter_gamma);
toc;

fprintf('[mirrorDescentWrapper] Objective = %e\n',finalLpDualObj);
%%%END function solve_hkl_hinge_shogun_pratik

function [zetaFake] = calculateZetaFake(affs,rho,weights,eta)
p =length(eta);
mat = zeros(p,p);
for i=1:p
    mat(i,affs{i}) = 1;
end
dffs = cell(p,1);
for i=1:p
    dffs{i} = find(mat(:,i)==1);
end

weightRho = weights.^(rho/(1-rho));
zetaFake = zeros(p,1);
etaFake = weightRho.*eta;
for i=1:p
    zetaFake(i,1) = norm(etaFake(dffs{i}),1-rho);
end


function [answer flag] = calculateSumForDesc(aTQa_smo,affs,rho,weights,eta,zetaFake)

p =length(eta);
% etaRho = eta.^rho;
% weightRho = weights.^rho;

if (rho ~= 2)
    rhobar = rho/(2-rho);% MKL norm in primal
    rhobarstar = rhobar/(rhobar-1);
else
    rhobar = inf;% MKL norm in primal
    rhobarstar = 1;
end

zeroEta = (eta==0);
% keyboard;
aTQa_smo = aTQa_smo./(zetaFake.^(1/rhobarstar));% aTQa_smo passed to the function has zetaFake in it.
aTKaZeta = (aTQa_smo.^(0.5/(rho-1))).*(zetaFake);

answer = zeros(p,1);
for i=1:p
    if zeroEta(i)
        answer(i) = sum( (aTKaZeta(affs{i})*weights(i)/eta(i)).^rho );
    else
        answer(i) = sum( (aTKaZeta(affs{i})*weights(i)/eta(i)).^rho );
    end
end
flag = 0;
if any( isnan(answer) | isinf(answer) )
    flag = 1;
    fprintf('[mirrorDescentWrapper] eta is zero\n');
    keyboard;
end

