function  [alphaFinal,dualObj,aTQa,b_t] = smo_lpMKL_reg_multiTask_K_mu_MKL(K_tr,nodeIdx_temp,zeta,hull,Z,alpha_alpha,C,L_p,details,tube_eps,mu)

% lpmkl(K1(:,:,nonZero),ones(length(nonZero),1),1/lambda,mkl_norm,Y,w1/n,w2/n);

% K_tr - 3-d matrix of size nxnxk. where n is number of training points and
% k is the number of kernels must be PD % YKY will be computed here depending on yKy
% Z - nx1 vector of outputs (all tasks included)
% alpha - alpha of last iteration, nx1 zero vector by default (all tasks included)
% C - regularization parameter
% L_p - p norm in the primal of lp-mkl
% details - details of the run. It contains
% details.numTrain - Tx1 vector number of training instances
% nodeIdx - numInstances x k where k is number of kernels


display = 0;

myEps = 1e-2;
TAU = 1e-12;
DTAU = 1e-5;
obj_threshold = 0.1;
diff_threshold = 1e-5;
myRidge = 1e-8;
MAXSMOITER = 1000000;

hull = hull-1;
singleTaskMuFactor = 1+1/mu;
T = details.T;
k = size(nodeIdx_temp,2);% number of nodes kernels
kern = size(K_tr,3);% number of types of kernels

numTrain = details.numTrain;
n_org = sum(numTrain);
n = 2*n_org;
begIdx = [1;cumsum(numTrain(1:end-1))+1];
endIdx = cumsum(numTrain(1:end));

wpos = ones(T,1);
wneg = ones(T,1);

% Norms
if (L_p == 1)
    L_q = inf;
else
    L_q = L_p/(L_p -1);
end

zetaRep = repmat(zeta,kern,1);
Q = zeros(2*size(K_tr,1),2*size(K_tr,2),size(K_tr,3));
for kk = 1:kern
    Ktemp = K_tr(:,:,kk);
    Q(:,:,kk) = [[Ktemp, -Ktemp];[-Ktemp', Ktemp]];
    Q(:,:,kk) = Q(:,:,kk) + myRidge*eye(size(Q,1));
end

nodeIdx = [nodeIdx_temp;nodeIdx_temp];
Y = [ones(n_org,1);-ones(n_org,1)];
p = -[(Z-tube_eps);(-Z-tube_eps)];


if (isempty(alpha_alpha))
    % cold start
    alpha = zeros(n,1);
    aTQa = zeros(k*kern,1);
    pTalpha = 0;
    T1 = 0;
    denom=0;
    numon = zeros(k*kern,1);
    obj = 0;
    d = zeros(k*kern,1);
    Qalpha = zeros(n,k*kern);
else
    % warm start - for alpha & aTQa
    alpha = [(alpha_alpha+abs(alpha_alpha))/2;(-alpha_alpha+abs(alpha_alpha))/2];
    pTalpha = p'*alpha;
    aTQa = zeros(k*kern,1);
    Qalpha = zeros(n,k*kern);
    for kk=1:kern
        for w=1:k
            index = (kk-1)*k+w;
            numTask = sum(hull(w,:)==1);
            if numTask == 1
                Qalpha(nodeIdx(:,w),index) = ( Q(nodeIdx(:,w),nodeIdx(:,w),kk)*alpha(nodeIdx(:,w)) )*(zeta(w)/singleTaskMuFactor);
                Qalpha(not(nodeIdx(:,w)),index) = alpha(not(nodeIdx(:,w)))*zeta(w)*myRidge;
                aTQa(index) = alpha'*Qalpha(:,index);
            else
                Qalpha(nodeIdx(:,w),index) = ( Q(nodeIdx(:,w),nodeIdx(:,w),kk)*alpha(nodeIdx(:,w)) )*zeta(w);
                Qalpha(not(nodeIdx(:,w)),index) = alpha(not(nodeIdx(:,w)))*zeta(w)*myRidge;
                aTQa(index) = alpha'*Qalpha(:,index);
            end
        end
    end
%     [d normTerm] = calObj(aTQa,L_q);% d: kx1, normTerm: 1x1
    
    T1 = norm(aTQa,L_q);
    denom = T1^(L_q - 1);
    numon = aTQa.^(L_q - 1);
    d = numon/denom;% kx1 vector
    obj = 0.5*T1;
    obj = obj + pTalpha;

%     Qalpha = shiftdim(alpha'*Q,1);% vector nxk
end

get_C = zeros(n,1);
idx_pos = (Y==1);
idx_neg = (Y==-1);
idx_t1 = cell(T,1);
idx_t2 = cell(T,1);
for t=1:T
    idx_temp = (begIdx(t):endIdx(t));
    idx = [idx_temp,idx_temp+n_org];
    idx_t1{t} = idx_temp;
    idx_t2{t} = idx_temp+n_org;
    get_C(idx(idx_pos(idx))) = C*wpos(t);
    get_C(idx(idx_neg(idx))) = C*wneg(t);
end
I_low = ((alpha>0)&(Y == 1)) | ((alpha<get_C)&(Y == -1)); % 1-0 vectors
I_up = ((alpha>0)&(Y == -1)) | ((alpha<get_C)&(Y == 1)); % 1-0 vectors

% SMO starts

% while KKT not satisfied
% choose the most violating pair
% update the pair

% i_old =0;j_old=0;

smoIter = 0;
nrIter = 0;
nrIterLocalMaxRand = 100;
nrRandStart = 0;
bisectIter = 0;
i_old = nan;
j_old = nan;
bestT_old = nan;
repeatCounter = 0;
repeatCounterGlobal = 0;
repeatCounterMax = 1000;
randStartCounter = 0;
gap = inf;
while (true)
    smoIter = smoIter+1;

    if display
        fprintf('SMO iteration = %d\n',smoIter);
    end

    if (mod(smoIter,1000)==0)
        fprintf('.');
%         keyboard;
    end
    if (mod(smoIter,50000)==0)
%         keyboard;
        fprintf('%f \n',obj);
%         disp(aTQa');
    end
    
    % CALCULATE gradient w.r.t. alpha - G - nx1 vector
    % Update G : g_from_Qalpha();
    G = Qalpha*d + p;
    
    % select working set
    %checkKKT()
    bestGap = -Inf;
    bestMaxIndex = nan;
    bestMinIndex = nan;
    bestT = nan;
    converge = true;
    yG = G.*Y;
    gapT = zeros(1,T);
    minIndexT = nan(1,T);
    maxIndexT = nan(1,T);
    for t=1:T
        iup = [(find(I_up(idx_t1{t}))+begIdx(t)-1);(find(I_up(idx_t2{t}))+begIdx(t)-1+n_org)];
        ilow = [(find(I_low(idx_t1{t}))+begIdx(t)-1);(find(I_low(idx_t2{t}))+begIdx(t)-1+n_org)];
        [maxVal max_i] = max( -yG(iup) );% max of -ve quantity = min of +ve quantity
        maxIndex = iup(max_i);
        [minVal min_j] = min( -yG(ilow) );% min of -ve quantity = max of -ve quantity
        minIndex = ilow(min_j);
        gap = (maxVal - minVal);
        gapT(t) = gap;
        minIndexT(t) = minIndex;
        maxIndexT(t) = maxIndex;
        if (gap > myEps)
            if gap > bestGap
                bestGap = gap;
                bestMaxIndex = maxIndex;
                bestMinIndex = minIndex;
                bestT = t;
                converge = false;
            end
        end
    end

    if converge
        break;% SMO converged
    elseif smoIter> MAXSMOITER
        fprintf('[smo_lpMKL_reg_multiTask_K_mu_MKL] - termination SMO - smoiter more than MAXSMOITER\n');
        break;
    end
    % main code
    % WSS1, i,j belong to same Task
    i = bestMaxIndex;% I_up
    j = bestMinIndex;% I_low
    
    if i==i_old && j==j_old && bestT==bestT_old
        repeatCounter = repeatCounter+1;
        repeatCounterGlobal = repeatCounterGlobal+1;
        if (repeatCounter > repeatCounterMax)
            fprintf('*');
            violatingTask = find(gapT>myEps);
            while length(violatingTask)>1
                bestT1 = violatingTask(randi(length(violatingTask),1));
                if bestT1 ~= bestT
                    bestT = bestT1;
                    break;
                end
            end
            i = maxIndexT(bestT);
            j = minIndexT(bestT);
            if length(violatingTask)>1
                randStartCounter = randStartCounter+1;
            end
            repeatCounter = 0;
        end
    else
        repeatCounter = 0;
    end
    i_old=i;j_old=j;bestT_old=bestT;
    
    Q_i = squeeze(Q(:,i,:));
    Q_j = squeeze(Q(:,j,:));
    C_i = get_C(i);
    C_j = get_C(j);
    old_alpha_i=alpha(i);
    old_alpha_j=alpha(j);

    %Find a direction of descent 
    rr_help = 2*myRidge*ones(1,k,kern);
    presentNodes = find(hull(:,bestT));
    absentNodes = find(1-hull(:,bestT));
    singleTaskNode = presentNodes(sum(hull(presentNodes,:),2)==1);
    if Y(i)==Y(j)
        if G(i)-G(j) < G(j) - G(i)
            dir_i=1;
            dir_j=-1;
            dir_grad=G(i) - G(j);
            dir_grad_offset=p(i) - p(j);
            eta_max=min(C_i - alpha(i), alpha(j));
        else
            dir_i=-1;
            dir_j=1;
            dir_grad=G(j) - G(i);
            dir_grad_offset=p(j) - p(i);
            eta_max=min(alpha(i), C_j - alpha(j));
        end
        rr_help(1,presentNodes,:)=repmat(Q(i,i,:) + Q(j,j,:) - 2*Q(i,j,:),1,length(presentNodes));
    else
        if G(i)+G(j) < -G(j) - G(i)
            dir_i=1;
            dir_j=1;
            dir_grad=G(i) + G(j);
            dir_grad_offset=p(i) + p(j);
            eta_max=min(C_i - alpha(i), C_j - alpha(j));
        else
            dir_i=-1;
            dir_j=-1;
            dir_grad=-G(i) - G(j);
            dir_grad_offset=-p(j) - p(i);
            eta_max=min(alpha(i), alpha(j));
        end
        rr_help(1,presentNodes,:)=repmat(Q(i,i,:) + Q(j,j,:) + 2*Q(i,j,:),1,length(presentNodes));
    end
    rr_help(1,singleTaskNode,:) = rr_help(1,singleTaskNode,:)/singleTaskMuFactor;
    rr = rr_help(:)';
    rr = rr.*zetaRep';
    
    % Compute directional Hessian
    % Newton - raphson
    nrIter = nrIter+1;% eta = 0 (eta has been used instead of delta)

    % obtain dir_hess
    if (T1==0)
        dir_hess = 0;
    else
        H1 = (1-L_q)*(4*(dir_grad-dir_grad_offset)^2)/T1;
        TH1 = (dir_i*Qalpha(i,:) + dir_j*Qalpha(j,:));% b_k
        H2 = (2*rr*d + 4*(L_q-1)*(TH1.^2)*(d./aTQa));% 4 is due to square(2b_k)
        dir_hess = 0.5*(H1+H2);
        if (dir_hess < 0)
            dir_hess = TAU;
%             fprintf('\n-ve hessian\n');
%             keyboard;
        end
    end
    
    if dir_hess <= 0
        dir_hess=TAU;
    end
    
    old_obj = obj;
    old_dir_grad = dir_grad;
    eta_min = 0;
    eta = -dir_grad/dir_hess;
    if(eta > eta_max)
        eta = eta_max;
    elseif eta < eta_min
        eta = 0.5*(eta_max + eta_min);
    end
    nrIterLocal = 0;
    while true
%         keyboard;
        oldAlpha = alpha;
%         counter = counter+1;
%         objVec(counter) = obj;
        new_alpha_i=old_alpha_i + eta*dir_i;
        new_alpha_j=old_alpha_j + eta*dir_j;
        change_i=new_alpha_i - alpha(i);
        change_j=new_alpha_j - alpha(j);      
        alpha(i)=new_alpha_i;
        alpha(j)=new_alpha_j;
        
        % update_Qalpha - i.e. aTQa
        r1_help = myRidge*ones(k,1,kern);
        r1_help(presentNodes,1,:) = repmat(Q(i,i,:),length(presentNodes),1);% % 10Sep2010 +myRidge;
        r1_help(singleTaskNode,1,:) = r1_help(singleTaskNode,1,:)/singleTaskMuFactor;
        r1 = r1_help(:);

        r2_help = myRidge*ones(k,1,kern);
        r2_help(presentNodes,1,:) = repmat(Q(j,j,:),length(presentNodes),1);% % 10Sep2010 +myRidge;
        r2_help(singleTaskNode,1,:) = r2_help(singleTaskNode,1,:)/singleTaskMuFactor;
        r2 = r2_help(:);
        
        r3_help = zeros(k,1,kern);
        r3_help(presentNodes,1,:) = repmat(Q(i,j,:),length(presentNodes),1);
        r3_help(singleTaskNode,1,:) = r3_help(singleTaskNode,1,:)/singleTaskMuFactor;
        r3 = r3_help(:);

        aTQa = aTQa + ( r1*change_i*change_i + r2*change_j*change_j + 2*r3*change_i*change_j ).*zetaRep +...
            2*( Qalpha(i,:)*change_i+Qalpha(j,:)*change_j)'; %a*delta^2 + 2*b*delta;
        if any(aTQa < 0)
            fprintf('[smo_lpMKL_reg_multiTask_K_mu_MKL] -ve aTQa\n');
            if all(aTQa(aTQa<0)>-1e-9)
                fprintf('[smo_lpMKL_reg_multiTask_K_mu_MKL] -ve aTQa within tolerable limits of %e. Making them +ve %e\n',1e-9,1e-8);
                aTQa(aTQa<0) = 1e-8;
            else
                keyboard;
            end
        end
        
        % update_Qalpha
        if ~isempty(absentNodes)
            % for absentNodes
            index_help = bsxfun(@plus,absentNodes,(0:kern-1)*k);
            index = index_help(:);
            Qalpha(i,index) = Qalpha(i,index) + myRidge*change_i*zetaRep(index)';
            Qalpha(j,index) = Qalpha(j,index) + myRidge*change_j*zetaRep(index)';
        end
        for w=presentNodes'
            q1 = zeros(n,kern);
            q2 = zeros(n,kern);
            if w == singleTaskNode
                q1(nodeIdx(:,w),:) = Q_i(nodeIdx(:,w),:)*(change_i*zeta(w)/singleTaskMuFactor);
                q2(nodeIdx(:,w),:) = Q_j(nodeIdx(:,w),:)*(change_j*zeta(w)/singleTaskMuFactor);
            else
                q1(nodeIdx(:,w),:) = Q_i(nodeIdx(:,w),:)*(change_i*zeta(w));
                q2(nodeIdx(:,w),:) = Q_j(nodeIdx(:,w),:)*(change_j*zeta(w));
            end
            index = bsxfun(@plus,w,(0:kern-1)*k);
            Qalpha(:,index) = Qalpha(:,index) + q1 + q2;
        end
        
        pTalpha = pTalpha + p(i)*change_i + p(j)*change_j;
        T1 = norm(aTQa,L_q);
        denom = T1^(L_q - 1);
        numon = aTQa.^(L_q - 1);
        d = numon/denom;% kx1 vector

        obj = 0.5*T1;
        obj = obj + pTalpha;
        
        % Compute new directional gradient
        dir_grad = dir_grad_offset;
        TH1 = (dir_i*Qalpha(i,:) + dir_j*Qalpha(j,:));% b_k
        largeOnes = d>DTAU;
        dir_grad = dir_grad + TH1(largeOnes)*d(largeOnes);

        if ( obj < old_obj + obj_threshold*eta*old_dir_grad ||...
                abs(eta_max - eta_min) < diff_threshold ||...
                abs(dir_grad) < diff_threshold ||...
                (eta == eta_max && dir_grad < 0) ||...
                (eta == eta_min && dir_grad > 0) )
            % Small problem convergence
            break;
        end
        if(dir_grad > 0)
            eta_max=eta;
        elseif(eta >= eta_min+TAU)
            eta_min=eta;
        end
        
        % Try a Newton-Raphson step
        % Compute directional Hessian
        
        % obtain dir_hess
        H1 = (1-L_q)*(4*(dir_grad-dir_grad_offset)^2)/T1;
        H2 = (2*rr*d + 4*(L_q-1)*(TH1.^2)*(d./aTQa));
        dir_hess = 0.5*(H1+H2);

        if(dir_hess <= 0)
            dir_hess=TAU;
        end
        eta=eta - dir_grad/dir_hess;
        % If the result is outside the brackets, use bisection
        if(eta > eta_max || eta < eta_min)
            bisectIter = bisectIter+1;
            eta = 0.5*(eta_max + eta_min);
        end
        nrIter = nrIter+1;
        nrIterLocal = nrIterLocal+1;
        if mod(nrIterLocal,nrIterLocalMaxRand)==0
            nrRandStart = nrRandStart+1;
            eta = 0.5*(eta_max + eta_min);
        end
    end

    % update gradients w.r.t. alpha using d, Qalpha
    
%     newtonRaphsonIter(smoIter,1) = nrIter;
    
    if (display==1)
        fprintf('[smo_lpMKL_reg_multiTask_K_mu_MKL] Value of delta: %f\n',eta);
    end
    
    % update I_up I_low accordingly
    if (alpha(i)<C_i && Y(i)==1) || (alpha(i)>0 && Y(i)==-1)
        I_up(i) = 1;
    else
        I_up(i) = 0;
    end
    if (alpha(i)<C_i && Y(i)==-1) || (alpha(i)>0 && Y(i)==1)
        I_low(i) = 1;
    else
        I_low(i) = 0;
    end
    if (alpha(j)<C_j && Y(j)==1) || (alpha(j)>0 && Y(j)==-1)
        I_up(j) = 1;
    else
        I_up(j) = 0;
    end
    if (alpha(j)<C_j && Y(j)==-1) || (alpha(j)>0 && Y(j)==1)
        I_low(j) = 1;
    else
        I_low(j) = 0;
    end
end

% calculate b_t
b_t = zeros(T,1);
yG = G.*Y;
nr_free = alpha > 0 & alpha < get_C;
for t=1:T
    nr_free_task = nr_free([idx_t1{t},idx_t2{t}]);
    yG_task = yG([idx_t1{t},idx_t2{t}]);
    if any(nr_free_task)
        b_t(t) = sum(yG_task(nr_free_task))/sum(nr_free_task);
    else
        iup = [(find(I_up(idx_t1{t}))+begIdx(t)-1);(find(I_up(idx_t2{t}))+begIdx(t)-1+n_org)];
        ilow = [(find(I_low(idx_t1{t}))+begIdx(t)-1);(find(I_low(idx_t2{t}))+begIdx(t)-1+n_org)];
        maxVal = min(yG(iup));
        minVal = max(yG(ilow));
        b_t(t) = (minVal+maxVal)/2;
    end
end
b_t = -b_t;% following the decision model of libsvm
% calculate final objective
fprintf('[smo_lpMKL_reg_multiTask_K_mu_MKL] gap: %e, smoIter: %d, nr: %d, bs: %d, randStart: %d, repeats: %d, nrRandStart: %d\n',max(gapT),smoIter,nrIter,...
    bisectIter,randStartCounter,repeatCounterGlobal,nrRandStart);
dualObj = -obj;
alphaFinal = alpha(1:n_org) - alpha((1:n_org)+n_org);
