function output = solve_htl(K_tr,Y,C,rho,data,dataset,mu,varargin)

% K_Tr : YKY
% global K_tr Y_tr;% store YKY

maxtoadd_suff = 500; % number of variables added after suff
myTolEta = 1e-10;
myTolTheta = 1e-8;
mingap = 1e-2;
mingapObj = 1e-6;
maxactive = 10000;
hull =[];
nodeIdx = logical([]);
eta_hull = [];
alphaStruct = [];
nonZeroNodes = nan;
inverted = false;
gap1 = nan;
converge = 0;% 0: not converge, 1: converge by sufficiency condition, 2: converge by rel change in objective, 3: all nodes inside hull
current_value = nan;
alphaCell = cell(1,1);
etaCell = cell(1,1);
thetaCell = cell(1,1);
zetaCell = cell(1,1);
objectiveValueArray = [];
activeSetIter = 0;

% av_change = 2;
% READ OPTIONAL PARAMETERS
args = varargin;
nargs = length(args);
for i=1:2:nargs
    switch args{i},
        case 'display',        display = args{i+1};
%         case 'mingap',        mingap = args{i+1};
        case 'hull',        hull = args{i+1};
        case 'alpha',        alphaStruct = args{i+1};
%         case 'b',               b = args{i+1};
        case 'eta_hull',             eta_hull = args{i+1};
%         case 'rhonorm',        rhonorm = args{i+1};
% 		case 'conflictMatrix',   	conflictMatrix = args{i+1};
        case 'maxactive',        maxactive = args{i+1};
		case 'av_change',		 av_change = args{i+1};            
%         case 'max_difference_active',        max_difference_active = args{i+1};
%         case 'maxtoadd_nece',        maxtoadd_nece = args{i+1};
        case 'maxtoadd_suff',        maxtoadd_suff = args{i+1};
%         case 'gapeta',        gapeta = args{i+1};
%         case 'conjgrad',        conjgrad = args{i+1};
        case 'inverted',        inverted = args{i+1};
%         case 'mltask',        mltask = args{i+1};
    end
end

% INITIALIZATION OF HULL WHEN NONE IS GIVEN
if inverted
    taskDenoter = 1;
else
    taskDenoter = 2;
end

if size(hull,1) > 2000
    maxtoadd_suff = 1000;
else
    maxtoadd_suff = 500;
end

if isempty(hull)
%     hull = ones(data.T,data.T)+eye(data.T);% single task nodes - NO TOP node
%          hull = perms_reps([1 2],[data.T-1,1]);
    if inverted
        % inverted
        hull = [];
        for v=0:1;%data.T-1;
            P = perms_reps([1 2],[data.T-v,v]);
            hull = [hull;P];
        end
    else
        % non inverted
        hull = [];
        for v=1:1;data.T;
            P = perms_reps([1 2],[data.T-v,v]);
            hull = [hull;P];
        end
    end
end

[data_hull,sources] = get_reduced_graph_weights_sources(data,hull);

if inverted
    ssum = sum(sources,2);
    idx = (ssum == 2*data.T);
    sources(idx,:) = [];
end
numTrain = dataset.numTrain;
n = sum(numTrain);
begIdx = [1;cumsum(numTrain(1:end-1))+1];
endIdx = cumsum(numTrain(1:end));
% strIdx = size(nodeIdx,2);
% stpIdx = size(data_hull.hull,1);
% for i=strIdx:stpIdx
%     toAdd = zeros(n,1);
%     presentTasks = find(data_hull.hull(i,:));
%     for t = presentTasks
%         toAdd(begIdx(t):endIdx(t)) = 1;
%     end
%     nodeIdx = [nodeIdx, logical(toAdd)];
% end
toAdd = false(n,size(data_hull.hull,1)-size(nodeIdx,2));
for t=1:dataset.T
    presentNodes = data_hull.hull((size(nodeIdx,2)+1:end),t)==taskDenoter;
    toAdd(begIdx(t):endIdx(t),presentNodes) = true;
end
nodeIdx = [nodeIdx, toAdd];
if isempty(eta_hull),
    eta_hull = ones(size(hull,1),1);
end

% % make sure normalization is correct
eta_hull = eta_hull/( sum( eta_hull ));
%Solve small problem using MD+SMO
[alphaStruct,b_t,eta_hull,omegaSquared,objectiveValue,theta_hull,zeta_hull,aTQa] = ...
    mirrorDescentWrapper(K_tr,alphaStruct,Y,nodeIdx,C,rho,eta_hull,data_hull,dataset,inverted,mu);
oldObjectiveValue = objectiveValue;
activeSetIter = activeSetIter+1;
objectiveValueArray(activeSetIter) = objectiveValue;
alphaCell{activeSetIter} = alphaStruct;
etaCell{activeSetIter} = eta_hull;
thetaCell{activeSetIter} = theta_hull;
zetaCell{activeSetIter} = zeta_hull;

current_value = omegaSquared;

% SUFFICIENT CONDITION CHECK

fprintf('\nENTERING SUFFICIENT PHASE\n\n');
% sufficient_phase = 1;

if size(hull,1)>maxactive
    sufficient_phase=0;
else
    sufficient_phase = 1;
end

while  sufficient_phase > 0
	if size(hull,1) > 2000
	    maxtoadd_suff = 1000;
	else
	    maxtoadd_suff = 500;
	end
    if isempty(sources)
        % everything is selected: exit!
        sufficient_phase = Inf;
        converge = 3;
        break;
    end
    
    % compute sufficient condition
    tic;
    gapsuff = check_sufficient_gaps_efficient(sources,aTQa,data,inverted);
    fprintf('[solve_htl] check_sufficient_gaps_efficient Time: ');
    toc;
    
    gap1 = max(gapsuff)-current_value;
    output.gap1 = gap1;

    if isnan(gap1)
        fprintf('GAP is NAN\n');
        keyboard;
    end
    
    [a,b]=sort(-gapsuff);
    sources = sources(b,:);
    gapsuff = gapsuff(b);
    addhull_sources =  find(gapsuff >=  current_value*( 1+ mingap) );
    
    fprintf('[solve_ghtl] abs suff gap1 = %e, rel suff gap1 = %e violating sources = %d, current_value = %e\n', gap1, gap1/current_value, ...
        length(addhull_sources), current_value);
    
    addhull_sources = addhull_sources(1:min(maxtoadd_suff,length(addhull_sources)));

    % prepare new hull for next step
    if isempty(addhull_sources)
        sufficient_phase = 0;
        converge = 1;
        fprintf('\naddhull_sources EMPTY\n');
        break;
    end
    tic;
    [data_hull,sources] = update_reduced_graph_weights_sources(data,data_hull,sources,addhull_sources);% we can add conflict matrix later on
    fprintf('[solve_htl] update_reduced_graph_weights_sources Time: ');
    toc;
    if inverted
        ssum = sum(sources,2);
        idx = (ssum == 2*data.T);
        sources(idx,:) = [];
    end
    hull = data_hull.hull;
    % for nodeIdx
    numTrain = dataset.numTrain;
    n = sum(numTrain);
    begIdx = [1;cumsum(numTrain(1:end-1))+1];
    endIdx = cumsum(numTrain(1:end));
    toAdd = false(n,size(data_hull.hull,1)-size(nodeIdx,2));
    for t=1:dataset.T
        presentNodes = data_hull.hull((size(nodeIdx,2)+1:end),t)==taskDenoter;
        toAdd(begIdx(t):endIdx(t),presentNodes) = true;
    end
    nodeIdx = [nodeIdx, toAdd];
    eta_hull = [eta_hull; zeros(length(addhull_sources),1) ];
    
    %Solve small problem using MD+SMO
    [alphaStruct,b_t,eta_hull,omegaSquared,objectiveValue,theta_hull,zeta_hull,aTQa] = ...
        mirrorDescentWrapper(K_tr,alphaStruct,Y,nodeIdx,C,rho,eta_hull,data_hull,dataset,inverted,mu);

    activeSetIter = activeSetIter+1;
    objectiveValueArray(activeSetIter) = objectiveValue;
    alphaCell{activeSetIter} = alphaStruct;
    etaCell{activeSetIter} = eta_hull;
    thetaCell{activeSetIter} = theta_hull;
    zetaCell{activeSetIter} = zeta_hull;
    relDiff = (oldObjectiveValue - objectiveValue)/oldObjectiveValue;
    
    current_value = omegaSquared;
    if size(hull,1)>maxactive
        sufficient_phase=0;
    end
    
    if relDiff < mingapObj && size(hull,1) >= 2000
        fprintf('[solve_htl] Rel obj differenec sufficiently low: %e ... terminating the active set algo\n',relDiff);
        converge = 2;
        oldObjectiveValue = objectiveValue;
        break;
    else
        oldObjectiveValue = objectiveValue;
    end
end

fprintf('\n[solve_htl] Cleaning Phase begins\n');
% fprintf('\n[solve_htl] NO Cleaning Phase \n');
output.afterSuffObj = objectiveValue;
nodesAfterSufficiency = length(zeta_hull);

active = find(eta_hull > myTolEta);
% % compute hull with given set of active nodes
activeHull = indexActiveHull(active,data_hull.affinity,inverted);
hull = hull(activeHull,:);
data_hull = get_reduced_graph_weights_sources(data,hull);
nodeIdx = nodeIdx(:,activeHull);
eta_hull = eta_hull(activeHull);
eta_hull = eta_hull/sum(eta_hull);
% % solve small problem and get new values with this hull - MD+SMO
[alphaStruct,b_t,eta_hull,omegaSquared,objectiveValue,theta_hull,zeta_hull,aTQa] = ...
    mirrorDescentWrapper(K_tr,alphaStruct,Y,nodeIdx,C,rho,eta_hull,data_hull,dataset,inverted,mu,'final');

activeSetIter = activeSetIter+1;
objectiveValueArray(activeSetIter) = objectiveValue;
alphaCell{activeSetIter} = alphaStruct;
etaCell{activeSetIter} = eta_hull;
thetaCell{activeSetIter} = theta_hull;
zetaCell{activeSetIter} = zeta_hull;
relDiff1 = (oldObjectiveValue - objectiveValue)/oldObjectiveValue;
if abs(relDiff1) > 1e-3
    fprintf('[solve_htl] WARNING\n');
end
if abs(relDiff1) > 1e-2
    fprintf('[solve_htl] MAJOR WARNING\n');
end
fprintf('[solve_htl] rel obj diff after eta-sparsity: %e\n',relDiff1);
oldObjectiveValue = objectiveValue;

% % now etas are fixed
output.afterEtaObj = objectiveValue;
output.finalObj = objectiveValue;
% % get sparsity thru theta (with fixed eta)
if (rho ~= 2)
    rhobar = rho/(2-rho);
    active = zeros(size(theta_hull,1),1);
    active(theta_hull.^rhobar > myTolTheta) = 1;
    theta_hull = theta_hull.*active;
    theta_hull = theta_hull/norm(theta_hull,rhobar);
    % recompute alpha thru MTL-SVM (as theta is given)
    [alphaStruct, b_t, objectiveValue] = simpleSVM(K_tr,alphaStruct,Y,C,zeta_hull,theta_hull,data_hull,dataset,inverted,mu);
    % final sparsity = # of nodes in nodesAfterSufficiency with theta = 0
    nonZeroNodes = sum(theta_hull > 0);
    
    output.finalObj = objectiveValue;
    relDiff2 = (oldObjectiveValue - objectiveValue)/oldObjectiveValue;
    if abs(relDiff2) > 1e-3
        fprintf('[solve_htl] WARNING\n');
    end
    if abs(relDiff2) > 1e-2
        fprintf('[solve_htl] MAJOR WARNING\n');
    end
    fprintf('[solve_htl] rel obj diff after theta-sparsity: %e\n',relDiff2);
else
    fprintf('[solve_htl] rho=2,p=2 NOT implemented!!\n');
    keyboard;
    % final sparsity = # of nodes in nodesAfterSufficiency with eta = 0(rhonorm = 2 case)
    output.finalObj = objectiveValue;
    nonZeroNodes = sum(eta_hull > 0);
    fprintf('rho=2 - no theta-sparsity\n');
end

fprintf('[solve_htl] Non-zero kernels after cleaning: %d\n\n', nonZeroNodes);
nodeNumbers = [nodesAfterSufficiency, length(eta_hull), nonZeroNodes]';
output.nodeNumbers = nodeNumbers;

output.alphaStruct = alphaStruct;
output.b_t = b_t;
output.eta_hull = eta_hull;
output.theta_hull = theta_hull;
output.zeta_hull = zeta_hull;
output.data_hull = data_hull;
output.hull = hull;
output.absgap = gap1;
output.relgap = gap1/current_value;
output.current_value = current_value;
output.converge = converge;% 0: not converge, 1: converge by sufficiency condition, 2: converge by rel change in objective, 3: all nodes inside hull

output.activeSetIter = activeSetIter;
output.alphaCell = alphaCell;
output.etaCell = etaCell;
output.thetaCell = thetaCell;
output.zetaCell = zetaCell;
output.objectiveValueArray = objectiveValueArray;

% keyboard;

function gapsuff = check_sufficient_gaps_efficient(sources,aTQa,data,inverted)
% aTKSa - Tx1 vector where KS = sum of kernels

gapsuff = zeros(size(sources,1),1);
T = size(sources,2);
beta = data.weightA;
onePlueBetaSqInv = 1/(1+beta)^2;
onePlusOnePlueBetaSqInv = 1+onePlueBetaSqInv;
if ~inverted
    for u=1:size(sources,1)
        presentTasks = find(sources(u,:)==2);% row vector
        absentTasks = find(sources(u,:)==1);% row vector
        depth = length(presentTasks);
        n = T-depth;
        z = 1;
        % diagonal elements of presentTasks - presentTasks

        % T_i,T_j where both belong to presentTasks
        if (depth>1)
            pairs = combntns(presentTasks,2);
            index1 = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
            if (n>=1)
                pairs = setprod(presentTasks,absentTasks);
                index2 = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
                if (n>=2)
                    pairs = combntns(absentTasks,2);
                    index3 = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
                else
                    index3 = [];
                end
            else
                index2 = [];
                index3 = [];
            end
        else
            index1 = [];
            pairs = setprod(presentTasks,absentTasks);
            index2 = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
            pairs = combntns(absentTasks,2);
            index3 = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
        end

        gapsuff(u) = ((z/beta)^(2*depth))*(onePlusOnePlueBetaSqInv)^(T-depth-2)*( sum(aTQa([presentTasks';index1]))*(onePlusOnePlueBetaSqInv)^2 +...
            sum(aTQa([absentTasks';index2]))*(onePlueBetaSqInv)^2*(onePlusOnePlueBetaSqInv) + sum(aTQa([index3]))*(onePlueBetaSqInv)^4 );
    end
else
    for u=1:size(sources,1)
        presentTasks = find(sources(u,:)==1);% row vector
        depth = length(presentTasks);
        if depth<1
            fprintf('[check_sufficient_gaps_efficient] WARNING: depth less than one\n');
            keyboard;
        end
        z = 1;
        if (depth>1)
            pairs = combntns(presentTasks,2);
            index1 = T + (pairs(:,1)-1).*(T-pairs(:,1)/2) + pairs(:,2) - pairs(:,1);
        else
            index1 = [];
        end
        gapsuff(u) = ((z/beta)^(2*depth))*(onePlusOnePlueBetaSqInv)^(depth-2)*( sum(aTQa(presentTasks))*(onePlusOnePlueBetaSqInv) + sum(aTQa(index1)) );
    end
end

